/*-
 * Copyright (c) 2013 Alistair Crooks <agc@NetBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/param.h>

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <md5.h>
#include <rmd160.h>
#include <sha1.h>
#include <sha2.h>

#include "crc32c.h"

#include "merkletree.h"

typedef char *(*digestfunc_t)(const uint8_t *, size_t, char *);

/* an algorithm which can be used */
typedef struct Alg {
	const char	*name;	/* algorithm name */
	uint64_t	 size;	/* # of chars in formatted output */
	digestfunc_t	 func;	/* digest func */
} Alg;

#define MERKLETREE_MAX_ROWS	16

/* a row of digests in a merkle tree */
typedef struct merkletree_row_t {
	uint64_t	 outc;		/* # of output bytes */
	uint64_t	 blocks;	/* # of blocks in this row */
	char		*out;		/* sums for this row */
} merkletree_row_t;

/* a tree of rows */
typedef struct Rows {
	uint64_t		 height;	/* # of row in tree */
	merkletree_row_t	 row[MERKLETREE_MAX_ROWS];	/* rows of digests in tree */
} Rows;

/* small function to call native md5 to give correct args */
static char *
indirect_MD5Data(const uint8_t *v, size_t len, char *out)
{
	return MD5Data(v, (unsigned)len, out);
}

/* small function to call native rmd160 to give correct args */
static char *
indirect_RMD160Data(const uint8_t *v, size_t len, char *out)
{
	return RMD160Data(v, (unsigned)len, out);
}

/* small function to call native crc32c to give correct args */
static char *
indirect_crc32cData(const uint8_t *v, size_t len, char *out)
{
	uint32_t	ret;

	ret = crc32c_data(v, len);
	snprintf(out, 8 + 1, "%08x", ret);
	return out;
}

/* concatenate rmd160 and sha1 (pkgsrc) */
static char *
cat_RMD160_SHA1_Data(const uint8_t *v, size_t len, char *out)
{
	RMD160Data(v, (unsigned)len, out);
	return SHA1Data(v, len, &out[40]);
}

/* concatenate md5 and sha256 */
static char *
cat_MD5_SHA256_Data(const uint8_t *v, size_t len, char *out)
{
	MD5Data(v, (unsigned)len, out);
	return SHA256_Data(v, len, &out[32]);
}

/* concatenate sha512 and sha256 */
static char *
cat_SHA512_SHA256_Data(const uint8_t *v, size_t len, char *out)
{
	SHA512_Data(v, len, out);
	return SHA256_Data(v, len, &out[128]);
}

/* table of supported algorithms */
static Alg	algs[] = {
	{	"sha512",		128,	SHA512_Data		},
	{	"sha256",		64,	SHA256_Data		},
	{	"sha1",			40,	SHA1Data		},
	{	"rmd160",		40,	indirect_RMD160Data	},
	{	"ripemd160",		40,	indirect_RMD160Data	},
	{	"md5",			32,	indirect_MD5Data	},
	{	"crc32c",		8,	indirect_crc32cData	},
	/* concatenated digests */
	{	"rmd160,sha1",		80,	cat_RMD160_SHA1_Data	},
	{	"md5,sha256",		96,	cat_MD5_SHA256_Data	},
	{	"sha512,sha256",	192,	cat_SHA512_SHA256_Data	},
	{	NULL,			0,	NULL			}
};

/* find a digest algorithm */
static Alg *
findalg(const char *name)
{
	Alg	*alg;

	for (alg = algs ; alg->name ; alg++) {
		if (strcasecmp(name, alg->name) == 0) {
			return alg;
		}
	}
	return NULL;
}

/* round up to correct number of blocks */
static inline uint64_t
numblocks(uint64_t size, uint64_t blocksize)
{
	return (size + (blocksize - 1)) / blocksize;
}

/* dynamically allocate the tree */
static int
allocate_tree(merkletree_t *tree, uint64_t size, size_t h)
{
	Rows	*rows = (Rows *)tree->rows;
	Alg	*alg = (Alg *)tree->alg;

	if (h == 0) {
		rows->row[h].outc = size;
	}
	rows->row[h].blocks = numblocks(size, tree->blocksize);
	rows->row[h].out = calloc(1, rows->row[h].blocks * alg->size);
	if (size > 0 && rows->row[h].blocks > 1) {
		return allocate_tree(tree, size / (tree->blocksize * alg->size), ++rows->height);
	}
	return (int)++rows->height;
}

/* calculate a row in a tree */
static int
calc_row(merkletree_t *tree, FILE *fp, char *v, size_t size, size_t h)
{
	size_t	 xfer;
	size_t	 cc;
	size_t	 b;
	Rows	*rows = (Rows *)tree->rows;
	Alg	*alg = (Alg *)tree->alg;

	rows->row[h].outc = 0;
	for (b = cc = 0 ; cc < size; b++, cc += xfer) {
		if (fp) {
			xfer = fread(v, 1, tree->blocksize, fp);
			(*alg->func)((const uint8_t *)v, xfer, &rows->row[h].out[rows->row[h].outc]);
		} else {
			xfer = MIN(size - cc, tree->blocksize);
			(*alg->func)((const uint8_t *)&v[cc], xfer, &rows->row[h].out[rows->row[h].outc]);
		}
		rows->row[h].outc += alg->size;
	}
	if (h < rows->height - 1) {
		return calc_row(tree, NULL, rows->row[h].out, rows->row[h].outc, h + 1);
	}
	return 1;
}

/* skip over initial stuff */
static inline char *
skip_to_sum(char *s)
{
	char	*ret;
	char	*eq;

	if ((eq = strchr(s, '=')) == NULL) {
		ret = s;
	} else {
		for (ret = eq + 1; *ret == ' ' ; ret++) {
		}
	}
	return ret;
}

/* get the col'th column */
static inline char *
get_column(char *s, size_t col)
{
	size_t	 i;
	char	*space;
	char	*p;

	for (space = s, i = 0 ; i < col ; i++, space = p + 1) {
		if ((p = strchr(space + 1, ' ')) == NULL) {
			return NULL;
		}
	}
	return space;
}

/* set up the tree */
static inline int
setup_tree(merkletree_t *tree, size_t size, const char *alg, size_t blocksize)
{
	if (tree == NULL || alg == NULL || blocksize == 0) {
		fprintf(stderr, "bad arg\n");
		return 0;
	}
	if ((tree->alg = findalg(alg)) == NULL) {
		fprintf(stderr, "unrecognised algorithm '%s'\n", alg);
		return 0;
	}
	tree->size = size;
	tree->blocksize = blocksize;
	tree->rows = calloc(1, sizeof(Rows));
	allocate_tree(tree, tree->size, 0);
	return 1;
}

/************************************************************************/

/* calculate a merkle tree on data */
int
merkletree_data(merkletree_t *tree, const void *v, size_t size, const char *alg, size_t blocksize)
{
	if (!setup_tree(tree, size, alg, blocksize)) {
		return 0;
	}
	return calc_row(tree, NULL, (char *)__UNCONST(v), tree->size, 0);
}

/* calculate a merkle tree on a file */
int
merkletree_file(merkletree_t *tree, const char *f, const char *alg, size_t blocksize)
{
	struct stat	 st;
	size_t		 size;
	uint8_t		*mapped;
	char		*alloc;
	FILE		*fp;
	int		 ret;

	if (tree == NULL || f == NULL || alg == NULL || blocksize == 0) {
		return 0;
	}
	if ((fp = fopen(f, "r")) == NULL) {
		fprintf(stderr, "can't open file '%s'\n", f);
		return 0;
	}
	ret = 0;
	fstat(fileno(fp), &st);
	size = st.st_size;
	alloc = NULL;
	if ((mapped = mmap(NULL, size, PROT_READ, MAP_SHARED, fileno(fp), 0)) == MAP_FAILED) {
		if ((alloc = calloc(1, blocksize)) == NULL) {
			goto done;
		}
		if (!setup_tree(tree, size, alg, blocksize)) {
			goto done;
		}
		ret = calc_row(tree, fp, alloc, tree->size, 0);
	} else {
		ret = merkletree_data(tree, mapped, size, alg, blocksize);
	}
done:
	if (mapped != MAP_FAILED) {
		munmap(mapped, tree->size);
	}
	if (alloc) {
		free(alloc);
	}
	fclose(fp);
	return ret;
}

/* format a merkle tree */
int
merkletree_snprintf(merkletree_t *tree, const char *f, char *buf, size_t size)
{
	size_t	 cc;
	size_t	 h;
	Rows	*rows;
	Alg	*alg;

	if (tree != NULL && buf != NULL && size > 0) {
		rows = (Rows *)tree->rows;
		alg = (Alg *)tree->alg;
		cc = 0;
		if (f == NULL) {
			f = "[stdin]";
		}
		cc = snprintf(buf, size,
			"merkle %s %" PRIu64 " %" PRIu64 " %" PRIu64 " (%s) = ",
			alg->name, rows->height, tree->size, tree->blocksize, f);
		for (h = 0 ; h < rows->height ; h++) {
			cc += snprintf(&buf[cc], size - cc, "%s", rows->row[h].out);
		}
		return (int)cc;
	}
	return 0;
}

/* format a merkle tree */
int
merkletree_asprintf(merkletree_t *tree, const char *f, char **buf)
{
	size_t	 size;
	size_t	 cc;
	size_t	 h;
	Rows	*rows = (Rows *)tree->rows;
	Alg	*alg = (Alg *)tree->alg;

	if (tree != NULL) {
		if (f == NULL) {
			f = "[stdin]";
		}
		for (size = 0, h = 0 ; h < rows->height ; h++) {
			size += strlen(rows->row[h].out);
		}
		size += strlen(f) + 100;
		if ((*buf = calloc(1, size)) == NULL) {
			return 0;
		}
		cc = snprintf(*buf, size,
			"merkle %s %" PRIu64 " %" PRIu64 " %" PRIu64 " (%s) = ",
			alg->name, rows->height, tree->size, tree->blocksize, f);
		for (h = 0 ; h < rows->height ; h++) {
			cc += snprintf(&(*buf)[cc], size - cc, "%s", rows->row[h].out);
		}
		return (int)cc;
	}
	return 0;
}

/* free resources */
void
merkletree_free(merkletree_t *tree)
{
	uint64_t	 i;
	Rows		*rows;

	if (tree != NULL) {
		rows = (Rows *)tree->rows;
		for (i = 0 ; i < rows->height ; i++) {
			free(rows->row[i].out);
		}
		free(tree->rows);
	}
}

/* higher level */

/* retrieve sum information from a sum file */
int
merkletree_sum_info(merkletree_t *tree, const char *f, char **buf)
{
	struct stat	 st;
	FILE		*fp;
	char		*alg;
	char		*p;
	int		 ret;

	if (tree == NULL || f == NULL || buf == NULL || *buf == NULL) {
		return 0;
	}
	memset(tree, 0x0, sizeof(*tree));
	if ((fp = fopen(f, "r")) == NULL) {
		fprintf(stderr, "can't open '%s'\n", f);
		return 0;
	}
	fstat(fileno(fp), &st);
	*buf = calloc(1, (size_t)(st.st_size + 1));
	ret = (int)fread(*buf, 1, (size_t)(st.st_size - 1), fp);
	fclose(fp);
	if ((alg = get_column(*buf, 1)) == NULL) {
		return 0;
	}
	if ((p = get_column(*buf, 2)) == NULL) {
		return 0;
	}
	snprintf(tree->algname, sizeof(tree->algname), "%.*s", (int)(p - alg) - 1, alg);
	if ((tree->alg = findalg(tree->algname)) == NULL) {
		fprintf(stderr, "bad alg name '%s'\n", tree->algname);
		return 0;
	}
	if ((p = get_column(*buf, 3)) == NULL) {
		return 0;
	}
	tree->size = (uint64_t)strtoull(p, NULL, 0);
	if ((p = get_column(*buf, 4)) == NULL) {
		return 0;
	}
	tree->blocksize = (uint64_t)strtoull(p, NULL, 0);
	return ret;
}

/* verify a merkletree sum output against a file in the file system */
int
merkletree_verify(const char *sumfile, const char *f)
{
	merkletree_t	 store;
	merkletree_t	 calc;
	size_t		 off;
	size_t		 len;
	size_t		 i;
	char		*calcsum;
	char		*stored;
	char		*sum[2];
	Alg		*alg;

	if (sumfile == NULL || f == NULL) {
		return 1;
	}
	memset(&calc, 0x0, sizeof(calc));
	memset(&store, 0x0, sizeof(store));
	merkletree_sum_info(&store, sumfile, &stored);
	merkletree_file(&calc, f, store.algname, store.blocksize);
	merkletree_asprintf(&calc, f, &calcsum);
	alg = (Alg *)store.alg;
	sum[0] = skip_to_sum(calcsum);
	sum[1] = skip_to_sum(stored);
	if ((len = strlen(sum[0])) == strlen(sum[1]) && memcmp(sum[0], sum[1], len) == 0) {
		return 1;
	}
	for (off = 0, i = 0 ; off < store.size ; i += alg->size, off += store.blocksize) {
		if (memcmp(&sum[0][i], &sum[1][i], alg->size) != 0) {
			printf("[%" PRIu64 "] %" PRIu64 ", %" PRIu64 "\n",
				off / store.blocksize, off, off + store.blocksize - 1);
		}
	}
	return 0;
}
